/*
 *  Copyright (c) 1992-2012 Matisse Software Inc. All Rights Reserved.
 *
 *  This file (in both binary and source code form) is subject to the
 *  Supplemental License Terms governing use, modification and redistribution
 *  of Sample Code accompanying the Matisse(R) software.
 *
 */

// the default Matisse C++ header
#include "matisseCXX.h"

#include "company.h"

#include <iostream>
#include <sstream>
#include <iomanip>

using namespace std;

// namespaces containing Matisse classes
using namespace matisse;

// namespaces containing the Schema classes
using namespace examples::cxx_examples::chaps_10_11;

// How to delete objects?
// Often it is easiest just to use SQL, e.g. DELETE * FROM EMPLOYEE
// or DELETE * FROM EMPLOYEE WHERE ...", but it can be accomplished
// in client code.

// This application can be used in several ways to delete all the
// objects in the database or a range of objects.
//
// 1) Delete all the objects:  just pass the host and database args.
// The objects are deleted in groups of 100 by iterating through the
// instances, then the entire array removed at once.
//
// 2) Deleate a single object:  pass host, database, and Employee ID.
//
// 3) Delete a range of objects:  pass host, database, and the 
// first and last Employee IDs in the range to be deleted.  In this
// case, we loop-preload the entire array of objects into the client
// cache, then remove them individually.
//
// 4) Delete alternate objects in a range of objects, starting with
// the first:  pass the same four arguments as for (3), plus a fifth
// argument (the value is not checked, so any other value will work).
//
// The code for each of these four alternatives is optimized for the
// particular task.
//
// NOTE:  Before running delete, run load to generate sample data.

int main(int ac, char **av)
{
  if (ac < 3)
  {
    std::cerr << "Usage: " << av[0]
              << " <host> <db> [<idmin> <idmax> <trim>]"
              << endl;
    return -1;
  }

  try
  {
    MtDatabase db(av[1], av[2]);

    db.open();
    db.startTransaction();

    if (ac == 3)            // simpliest case, delete them all
    {
      int delCount = 100;
      std::cout << "Deleting all employees. (" << delCount
                << " at a time)" << std::endl;

      // get the iterator
      MtObjectIterator<Employee> eiter = Employee::instanceIterator(db);

      // An even more straighforward approach would be to use
      // Employee::getInstanceNumber(db) but that could possibly
      // create a very large (and unnecessary) array if instance
      // count was very high.
      while  (eiter.hasNext())
      {
        MtObjectArray<Employee> emps(eiter.next(delCount));

        std::cout << " removing " << emps.size() << " Employees"
                  << std::endl;
        emps.remove();
      }
      eiter.close();

    }
    else
    {
      // will be deleting a range
      bool trim = false;
      int fromId = ::atoi(av[3]);
      int toId = fromId;
      if (ac > 4)
        toId = ::atoi(av[4]);

      // existance of arg is indication to delete alternately
      if (ac > 5)
        trim = true;

      std::cout << "Deleting employees " << fromId << " to " << toId
                << (trim ? " (alternating)" : "")
                << std::endl;

      // get the iterator for entire range
      MtObjectIterator<Employee> eiter = Employee::employeeIdIterator(db,
                                                                      fromId, toId);

      // quick check to see if we have any instances to deal with
      if (eiter.hasNext())
      {
        if (trim)               // going to trim the list to delete
        {
          // get the whole array
          MtObjectArray<Employee> emps0(eiter.next(toId - fromId + 1));
          const MtOid *oids = emps0.oids();
          int oidCount = 0;

          // use auto ptr for safe cleanup
          auto_ptr<MtOid> tmpOids(new MtOid[emps0.size() / 2 + 1]);

          for (unsigned int j = 0; j < emps0.size(); j++)
          {
            if ((j % 2) == 0)
              *(tmpOids.get() + oidCount++) = oids[j];
                    
          }

          // now create new version of object array
          MtObjectArray<Employee> emps(db, oidCount, tmpOids.get());
          std::cout << " removing " << emps.size() << " Employees"
                    << std::endl;

          // could use MtObjectArray<>::remove as below, but
          // looping to check values, so load whole array first
          emps.load();
          int minId = emps[0].getId();
          int maxId = minId;
          for (unsigned int i = 0; i < emps.size(); i++)
          {
            if (emps[i].getId() < minId)
              minId = emps[i].getId();
            else if (emps[i].getId() > maxId)
              maxId = emps[i].getId();
            emps[i].remove();
          }
          std::cout << " min ID = " << minId
                    << ", max ID = " << maxId
                    << std::endl;
        }
        else            // not trimming, delete the specified range
        {
          // since we know id is unique, this is worst case
          int count = toId - fromId + 1;
          // grab the whole array
          MtObjectArray<Employee> emps(eiter.next(count));
          std::cout << " removing " << emps.size() << " Employees"
                    << std::endl;

          // could use MtObjectArray<>::remove as below, but
          // looping to check values, so load whole array first
          emps.load();
          int minId = emps[0].getId();
          int maxId = minId;
          for (unsigned int i = 0; i < emps.size(); i++)
          {
            if (emps[i].getId() < minId)
              minId = emps[i].getId();
            else if (emps[i].getId() > maxId)
              maxId = emps[i].getId();
            emps[i].remove();
          }
          std::cout << " min ID = " << minId
                    << ", max ID = " << maxId
                    << std::endl;
        }
      }
    }

    db.commit();
    db.close();

  } // end of try
  catch (MtException& mte)
  {
    std::cerr << "Matisse exception: " << mte << std::endl;
  }
  catch (std::exception& e)
  {
    std::cerr << "Standard exception: " << e.what() << std::endl;
  }
  catch (...)
  {
    std::cerr << "Unknown exception" << std::endl;
  }

  return 0;
}
