/*
 *  Copyright (c) 1992-2012 Matisse Software Inc. All Rights Reserved.
 *
 *  This file (in both binary and source code form) is subject to the
 *  Supplemental License Terms governing use, modification and redistribution
 *  of Sample Code accompanying the Matisse(R) software.
 *
 */

// the required Matisse C++ header
#include "matisseCXX.h"
// the headers for schema classes
#include "Manager.h"
#include "Employee.h"

// namespaces containing Matisse classes
using namespace matisse;
using namespace matisse::reflect;

// namespaces containing the Schema classes
using namespace examples::cxx_examples::chaps_6_7_8;


// A simple application that demonstrates how to access, modify, and navigate
// relationships.
int main(int ac, char **av)
{
  if (ac < 3)
  {
    std::cerr << "Usage: " << av[0] << " <host> <db>" << std::endl;
    return -1;
  }
  try
  {
    // In this example, there is a direct namespace mapping between 
    // the C++ namespace and the Schema namespace, so the default factory can handle it
    MtDatabase db(av[1], av[2]);

    matisse::MtTimestamp ts; // default is now
    matisse::MtNumeric num(48.3);

    db.open();
    db.startTransaction();

    // create a manager
    Manager& m1 = Manager::create(db);
    m1.setFirstName("J.C.");
    m1.setLastName("Dithers");
    m1.setHireDate(ts);
    m1.setSalary(num);
    // set the successor object for reportsTo (in this case it
    // refers to itself, i.e., this manager is the big boss at
    // the top of the reporting hierarchy)
    m1.setReportsTo(m1);

    // create another manager
    Manager& m2 = Manager::create(db);
    m2.setFirstName("Dagwood");
    m2.setLastName("Bumstead");
    m2.setHireDate(ts);
    num = 42.0;             // MtNumeric supports operator=(double)
    m2.setSalary(num);
    m2.setReportsTo(m1);    // the manager

    // create an employee
    Employee& e = Employee::create(db);
    e.setFirstName("Elmo");
    e.setLastName("Tuttle");
    ts = "2001-12-31";      // MtTimestamp supports operator=(char *)
    e.setHireDate(ts);
    num = 22.0;
    e.setSalary(num);
    e.setReportsTo(m2);

    // specify an assistant for each of the two managers
    m1.setAssistant(e);
    m2.setAssistant(e);

    // check the assistantOf from the employee
    std::cout << "\nInitial settings:" << std::endl;
    // get the entire list of sucessors (class Manager)
    MtObjectArray<Manager> list = e.getAssistantOf();
    // access list as an array using operator[]
    for (int i=0; i < (int)list.size(); i++)
      std::cout << "\t" << e.getFirstName() << " is "
                << list[i].getFirstName() << "'s assistant" << std::endl;

    // create a few extra persons
    Person& c1 = Person::create(db);
    c1.setFirstName("Alexander");
    c1.setLastName("Bumstead");
    
    Person& c2 = Person::create(db);
    c2.setFirstName("Cookie");
    c2.setLastName("Bumstead");

    // a std::vector of Type* can be converted to from
    // MtObjectArray<T> which can be used as argument to set a list
    // of successors
    std::vector<Person*> cVec(2);
    cVec[0] = &c1;
    cVec[1] = &c2;
    MtObjectArray<Person> children(db,cVec);
    m2.setChildren(children);


    // When accessing a relationship with single cardinality [1] or
    // [0,1], the access method returns a pointer to the object so
    // that it can be checked for NULL.
    // in this case getFather() returns Person*
    std::cout << "\nSet successors:" << std::endl;
    std::cout << "\t" << c1.getFirstName() << " is "
              << (c1.getFather())->getFirstName() << "'s child"
              << std::endl;
    std::cout << "\t" << c2.getFirstName() << " is "
              << (c2.getFather())->getFirstName() << "'s child"
              << std::endl;

    // convenient use of C++ overloaded operators
    std::cout << "\nCompare objects (use ==):" << std::endl;
    std::cout << "\tDo " << c1.getFirstName() << " and "
              << c2.getFirstName() <<  " have the same father (==)?     " 
              <<  (c1.getFather() == c2.getFather()) << std::endl;


    // the following code shows how to add and remove successors 
    Person& c3 = Person::create(db); 
    c3.setFirstName("Baby");
    c3.setLastName("Bumstead");

    // append to existing list
    m2.appendChildren(c3);
    
    std::cout << "\nAdd another successor:" << std::endl;
    std::cout << "\tNow " << m2.getFirstName() << " has " 
              << m2.getChildrenSize() << " children" << std::endl;

    // use of MtObjectArray<> for specifying a list to remove
    std::vector<Person*> rVec(2);
    rVec[0] = &c2;
    rVec[1] = &c3;
    MtObjectArray<Person> children2(db, rVec);
    // remove 
    m2.removeChildren(children2);

    // another signature supports a single removal 
    //  i.e.    m2.removeChildren(c2);
        
    unsigned int cnt =  m2.getChildrenSize();   
    std::cout << "\nRemove two successors:" << std::endl;
    std::cout << "\tNow " << m2.getFirstName() << " has " 
              << cnt << " children" << std::endl;
    // an alternative to get the relationship size 
    // but the C++ objects are loaded before you can get the count
    cnt =  m2.getChildren().size();
    std::cout << "\tNow " << m2.getFirstName() << " has " 
              << cnt << " children" << std::endl;

    // clear all the successors
    m2.clearChildren();
    std::cout << "\nClear successors:" << std::endl;
    std::cout << "\tNow " << m2.getFirstName() << " has " 
              << m2.getChildrenSize() << " children" << std::endl;

    // reset to original array
    m2.setChildren(children);

    // iterator
    std::cout << "\nAdd successors and iterate through children..."
              << std::endl;
    std::cout << "\t" << m2.getFirstName() << "'s children:" << std::endl;
    MtObjectIterator<Person> piter = m2.childrenIterator();
    while (piter.hasNext())
    {
      Person &p = piter.next();
      std::cout << "\t\t" << p.getFirstName() << std::endl;
    }

    //remove created objects
    m1.remove();
    m2.remove();
    e.remove();
    c1.remove();
    c2.remove();
    c3.remove();

    db.commit();
    db.close();
  }
  catch (MtException &e)
  {
    std::cerr << e << std::endl;
  }
  catch (...)
  {
    std::cerr << "Unknown exception" << std::endl;
  }
  return 0;
}
